(* Copyright (C) 2014, Daniel Wyckoff *)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)



Require Export BoolAlgBasics.
Require Import SetUtilities.
Require Export Image.
Require Import FunctionalExtensionality.
Require Export Basics.

Section infinite.
Variable B:Bool_Alg.
Let Bt := Btype (Bc B).
Variable It : Type.
Definition IS := Full_set It.


Definition Bcompi (p: It->Bt) : It->Bt :=
  fun (i:It) => -(p i).

Notation "-. x" := (Bcompi x) (at level 30).


Lemma doub_neg_ind : forall (p : It->Bt), -.(-.p) = p.
intros p.
unfold Bcompi.
cut ((fun i : It => - - p i) = (fun i : It => p i)).
intro h.
apply h.
apply functional_extensionality.
intro x.
rewrite doub_neg. reflexivity.
Qed.


Definition SupGen (Gt:Type) (S:Ensemble Gt) (p:Gt->Bt) (b:Bt) : Prop := 
sup (Im S p) b.

Definition InfGen (Gt:Type) (S:Ensemble Gt) (p:Gt->Bt) (b:Bt) : Prop := 
inf (Im S p) b.

Definition Sup (p: It->Bt) (b:Bt) : Prop :=
SupGen It IS p b.

Definition Inf (p: It->Bt) (b:Bt) : Prop :=
InfGen It IS p b.

Lemma comp_ind : 
  forall (p : It->Bt) (x:Bt), Im IS p (-x) <-> Im IS (-.p) x.
intros p x.
split.
(* left *)
intro h.
inversion h.
assert (h1: --x = -(p x0)).
rewrite H0. reflexivity.
assert (h2: --x = x).
rewrite doub_neg. reflexivity.
assert (h3: x = (-.p) x0).
unfold Bcompi. congruence.
apply Im_intro with (x := x0).
apply H. apply h3.
(* right *)
intro h4.
inversion h4.
unfold Bcompi in H0.
assert (h5: -x = --p x0).
rewrite H0. reflexivity.
assert (h6: --p x0 = p x0).
rewrite doub_neg. reflexivity.
assert (h7: p x0 = -x). congruence.
apply Im_intro with (x := x0).
assumption. rewrite h7. reflexivity.
Qed.

Lemma infnt_de_mor1 : forall (p : It->Bt) (b:Bt), Sup p b -> Inf (-.p) (-b).
intros p b h1.
unfold Sup in h1. unfold SupGen in h1. unfold sup in h1. elim h1.
intros h1l h1r.
unfold ub in h1l.
unfold Inf. unfold inf. split. unfold lb.
(* -b is a lower bound for -.p *)
intros t h2.
cut (le (-t) b).
rewrite doub_neg with (x := t) at 2.
apply mono_comp.
unfold In in h2.
rewrite <- comp_ind in h2.
apply h1l with (s := -t).
unfold In. assumption.
(* next subgoal: -b is the greatest lower bound *)
intros b'.
unfold lb.
unfold In.
intro h3.
rewrite doub_neg with (x := b').
apply mono_comp.
apply h1r with (b' := -b').
unfold ub.
unfold In.
intros s.
intro h4.
rewrite doub_neg in h4.
rewrite comp_ind in h4.
assert (h5: le b' (-s)).
apply h3 with (s := -s).
assumption.
rewrite doub_neg with (x := s).
apply mono_comp.
assumption.
Qed.

(*Dual of the previous of course.*)
Lemma infnt_de_mor2 : forall (p : It->Bt) (b : Bt), Inf p b -> Sup (-.p) (-b).
intros p b h1.
unfold Inf in h1. unfold InfGen in h1. unfold inf in h1. elim h1.
intros h1l h1r.
unfold lb in h1l.
unfold Sup. unfold sup. split.  unfold ub.
(* -b is an upper bound for -.p*)
intros t h2.
cut (le b (-t)).
rewrite doub_neg with (x :=t) at 2.
apply mono_comp.
unfold In in h2.
rewrite <- comp_ind in h2.
apply h1l with (s := -t).
unfold In.  assumption.
(* next suboal: -b is the greatest lower bound *)
intros b'.
unfold ub.
unfold In.
intro h3.
rewrite doub_neg with (x := b').
apply mono_comp.
apply h1r with (b' := -b').
unfold lb.
unfold In.
intros s.
intro h4.
rewrite doub_neg in h4.
rewrite comp_ind in h4.
assert (h5: le (-s) b').
apply h3 with (s := -s).
assumption.
rewrite doub_neg with (x := s).
apply mono_comp.
assumption.
Qed.

Lemma infnt_de_mor3 : forall (p : It->Bt) (b : Bt), Sup (-.p) b -> Inf p (-b).
intros p b h1.
rewrite <- doub_neg_ind with (p := p).
apply infnt_de_mor1 with (p := (-.p)).
assumption.
Qed.

Lemma infnt_de_mor4 : forall (p : It->Bt) (b : Bt), Inf (-.p) b -> Sup p (-b).
intros p b h1.
rewrite <- doub_neg_ind with (p := p).
apply infnt_de_mor2 with (p := (-.p)).
assumption.
Qed.

Definition complete : Prop := forall (p : It->Bt), (exists b_, Inf p b_) /\ 
                                                   (exists b', Sup p b').

Lemma complete_sup : (forall (p : It->Bt), (exists b', Sup p b')) -> complete.
intros h1.
unfold complete.
intro p. split.
assert (h2: exists b':Bt, Sup (-.p) b').
apply h1 with (p := -.p).
elim h2.
intros q.
intro h3.
clear h2.
assert (h4: Inf p (-q)).
apply infnt_de_mor3. assumption.
exists (-q).  assumption.
apply h1.
Qed.

(*Dual*)
Lemma complete_inf : (forall (p : It->Bt), (exists b_, Inf p b_)) -> complete.
intros h1.
unfold complete.
intro p. split.
apply h1.
assert (h2: exists b_:Bt, Inf (-.p) b_).
apply h1 with (p := -.p).
elim h2.
intros q.  intro h3. clear h2.
assert (h4: Sup p (-q)).
apply infnt_de_mor4. assumption.
exists (-q). assumption.
Qed.


Section infnt_assoc.
Variable Jt:Type.

Record Ind_Ind_Set : Type := 
{IIS: Ensemble (Ensemble It); 
 covers : (FamilyUnion IIS) = IS;
 non_empty_I : forall (S:(Ensemble It)), (In IIS S) -> S <> (Empty_set It);
 non_empty_II : exists S:(Ensemble It), In IIS S
}.

Variable II : Ind_Ind_Set.

Lemma inhbtd_IIS : forall (i:It), In (FamilyUnion (IIS II)) i.
intros i.
assert (h1:In IS i).
  unfold IS. apply Full_intro with (U:=It) (x:=i).
assert (h9: (FamilyUnion (IIS II)) = IS).
  apply covers.
rewrite h9.
assumption.
Qed.

Lemma infnt_assoc_sup : 
    (forall  (p: It -> Bt) (q:Jt->Bt), 
    (forall (S:Ensemble It), In (IIS II) S ->       
     (exists (j:Jt), SupGen It S p (q j))) -> 
    (forall (j:Jt), exists (S:Ensemble It), In (IIS II) S 
                          /\ SupGen It S p (q j)) ->
     (forall (q':Bt), SupGen Jt (Full_set Jt) q q' ->
      Sup p q')).
intros p q h1 h2' q'. intro h4.  
assert (h7: ub (Im IS p) q').
  unfold ub.
  intros b h8.
  unfold In in h8. inversion h8.
  assert (h5:forall (i:It), (exists (S1:(Ensemble It)), (In (IIS II) S1) /\
    In S1 i)).  
    intro i.
    assert (h6: In (FamilyUnion (IIS II)) i).
    apply inhbtd_IIS.
    inversion h6.
    exists S.
    tauto.
  assert (h10:exists S1 : Ensemble It, In (IIS II) S1 /\ In S1 x).
  apply h5.
  elim h10.
  intros S1 h.
  assert (h11: exists j : Jt, SupGen It S1 p (q j)).
  apply h1. apply h.
  elim h11. intros j h12.
  assert (h13: le b (q j)).
    apply h12.
    unfold In. 
    apply Im_intro with (x := x) (y := b).
    apply h.
    unfold In.
    assumption.
  assert (h14: le (q j) q').
    apply h4.
    unfold In. apply Im_intro with (x := j).
    apply Full_intro.  reflexivity.
  apply trans_le with (x := b) (y := (q j)) (z := q').
  assumption.  assumption.
  assert (h15: forall r:Bt, ub (Im IS p) r -> le q' r).
    intros r h16.
    assert (h17: forall (S: Ensemble It), In (IIS II) S ->
                 exists j : Jt, le (q j) r).
      intros S h30.
      assert (h31: exists j : Jt, SupGen It S p (q j)).
        apply h1. apply h30.
        elim h31.
        intros j h32.
    assert (h21: le (q j) r).
      apply h32. unfold ub.
        intros b' h22. unfold In in h22. inversion h22.        
        apply h16.           
      unfold In.
      assert (h23: In IS x).
        unfold IS.
      apply Full_intro.
      apply Im_intro with (x := x).
      assumption. assumption.
    exists j. assumption.
    assert (h24: ub (Im (Full_set Jt) q) r).
      unfold ub.  
      intros b h25. unfold In in h25. inversion h25. 
      assert (h26: exists S : Ensemble It,
        In (IIS II) S /\ SupGen It S p (q x)).
      apply h2'.
      elim h26. intros S h27.  
      elim h27. intros h27l h27r.
      rewrite <- H0 in h27r.
      apply h27r.  unfold ub.  
      intros b' h28. unfold In in h28. apply h16.
      inversion h28. rewrite H3. unfold In. apply Im_intro with (x:=x0).
      assert (h29: (FamilyUnion (IIS II)) = IS).
        apply covers.
      unfold IS. apply Full_intro. reflexivity. 
    apply h4. apply h24.
unfold Sup. unfold SupGen. unfold sup. split.
apply h7. apply h15.
Qed.

(*dual*)
Lemma infnt_assoc_inf : 
    (forall  (p: It -> Bt) (q:Jt->Bt), 
    (forall (S:Ensemble It), In (IIS II) S ->       
     (exists (j:Jt), InfGen It S p (q j))) -> 
    (forall (j:Jt), exists (S:Ensemble It), In (IIS II) S 
                          /\ InfGen It S p (q j)) ->
     (forall (q':Bt), InfGen Jt (Full_set Jt) q q' ->
      Inf p q')).
intros p q h1 h2' q' h4.  
assert (h7: lb (Im IS p) q').
  unfold lb.
  intros b h8.
  unfold In in h8. inversion h8.
  assert (h5:forall (i:It), (exists (S1:(Ensemble It)), (In (IIS II) S1) /\
    In S1 i)).  
    intro i.
    assert (h6: In (FamilyUnion (IIS II)) i).
    apply inhbtd_IIS.
    inversion h6.
    exists S.
    tauto.
  assert (h10:exists S1 : Ensemble It, In (IIS II) S1 /\ In S1 x).
  apply h5.
  elim h10.
  intros S1 h.
  assert (h11: exists j : Jt, InfGen It S1 p (q j)).
  apply h1. apply h.
  elim h11. intros j h12.
  assert (h13: le (q j) b).
    apply h12.
    unfold In. 
    apply Im_intro with (x := x) (y := b).
    apply h.
    unfold In.
    assumption.
  assert (h14: le q' (q j)).
    apply h4.
    unfold In. apply Im_intro with (x := j).
    apply Full_intro.  reflexivity.
  apply trans_le with (x := q') (y := (q j)) (z := b).
  assumption.  assumption.
  assert (h15: forall r:Bt, lb (Im IS p) r -> le r q').
    intros r h16.
    assert (h17: forall (S: Ensemble It), In (IIS II) S ->
                 exists j : Jt, le r (q j)).
      intros S h30.
      assert (h31: exists j : Jt, InfGen It S p (q j)).
        apply h1. apply h30.
        elim h31.
        intros j h32.
    assert (h21: le r (q j)).
      apply h32. unfold lb.
        intros b' h22. unfold In in h22. inversion h22.        
        apply h16.           
      unfold In.
      assert (h23: In IS x).
        unfold IS.
      apply Full_intro.
      apply Im_intro with (x := x).
      assumption. assumption.
    exists j. assumption.
    assert (h24: lb (Im (Full_set Jt) q) r).
      unfold lb.  
      intros b h25. unfold In in h25. inversion h25. 
      assert (h26: exists S : Ensemble It,
        In (IIS II) S /\ InfGen It S p (q x)).
      apply h2'.
      elim h26. intros S h27.  
      elim h27. intros h27l h27r.
      rewrite <- H0 in h27r.
      apply h27r.  unfold lb.  
      intros b' h28. unfold In in h28. apply h16.
      inversion h28. rewrite H3. unfold In. apply Im_intro with (x:=x0).
      assert (h29: (FamilyUnion (IIS II)) = IS).
        apply covers.
      unfold IS. apply Full_intro. reflexivity. 
    apply h4. apply h24.
unfold Inf. unfold InfGen. unfold inf. split.
apply h7. apply h15.
Qed.
End infnt_assoc.


Lemma infnt_distr_1_sup : forall {T:Type} (p:Bt) (q:T->Bt) (q':Bt), 
  let pq := (compose (fun x:Bt => p*x) q) in  
  SupGen _ (Full_set T) q q' -> SupGen _ (Full_set T) pq (p*q').
intros T p q q' pq h1.
assert (h2:ub (Im (Full_set T) pq) (p*q')).
  unfold ub.
  intro s. unfold In. intro h3. inversion h3. 
  unfold pq in H0. unfold compose in H0.
  rewrite H0.
  apply mono_prod. apply refl_le.
  apply h1. unfold In. apply Im_intro with (x:=x). apply H.
    reflexivity.
assert (h4: forall r:Bt, ub (Im (Full_set T) pq) r -> le (p*q') r).
 intros r h5.
 unfold ub in h5. 
 assert (h6: ub (Im (Full_set T) q) (r + (-p))).
   unfold ub. intros b h7.
   unfold In in h7. inversion h7.
   assert (h8: b = (p + (-p)) * b). 
     rewrite comp_sum.  rewrite comm_prod.  
     rewrite one_prod. reflexivity. 
   assert (h9: (p + (-p))*b = (p*b) + (-p)*b).
     apply dist_sum_r.
   assert (h10: b = (p*b) + (-p)*b).
     congruence.
   assert (h11:le (p*b) r).
     apply h5.
     unfold In.
     rewrite H0.
     apply Im_intro with (x := x).
     apply H.
     unfold pq.
     unfold compose. reflexivity.
   assert (h12:le (-p*b) (-p)).
     unfold le. apply eq_ord. 
     assert (h13: (-p)*b*(-p) = (-p)*(-p)*b).  
       rewrite <- assoc_prod.
       rewrite (comm_prod _ b (-p)). 
       rewrite assoc_prod.  reflexivity. 
     assert (h14: (-p)*(-p) = (-p)).
       apply idem_prod.
     rewrite h14 in h13.
     apply h13.
   assert (h15 : le (p*b + (-p*b)) (r + -p)).
     apply mono_sum.
   assumption.  assumption.
   rewrite  h10. assumption.
assert (h11: le q' (r+ (-p))).
  apply h1. assumption.
assert (h12: le p p).
  apply refl_le.
assert (h13: le (p*q') (p*(r+-p))).
  apply mono_prod. assumption. assumption.
assert (h14: p*(r+ -p) = p*r + p*-p).
  apply dist_sum.
assert (h15: p*r + p*-p = p*r).  
  rewrite comp_prod.  rewrite zero_sum.  reflexivity. 
assert (h16: le (p*r) r). unfold le.  apply eq_ord.
    assert (h18: r*r = r).  apply idem_prod.
    assert (h19: p*r*r = p*(r*r)). 
      rewrite assoc_prod. reflexivity.
    rewrite h18 in h19. assumption.
rewrite <- h15 in h16.
rewrite <- h14 in h16.
apply trans_le with (x := (p*q')) (y := (p*(r+-p))) (z:=r).
assumption. assumption.
unfold Sup. unfold SupGen. unfold sup. split.
assumption. assumption.
Qed.

Lemma infnt_distr_1_inf : forall {T:Type} (p:Bt) (q:T->Bt) (q':Bt), 
  let pq := (compose (fun x:Bt => p+x) q) in  
  InfGen _ (Full_set T) q q' -> InfGen _ (Full_set T) pq (p+q').
intros T p q q' pq h1.
assert (h2:lb (Im (Full_set T) pq) (p+q')).
  unfold lb.
  intro s. unfold In. intro h3. inversion h3. 
  unfold pq in H0. unfold compose in H0.
  rewrite H0.
  apply mono_sum. apply refl_le.
  apply h1. unfold In. apply Im_intro with (x:=x). apply H.
    reflexivity.
assert (h4: forall r:Bt, lb (Im (Full_set T) pq) r -> le r (p+q')).
 intros r h5.
 unfold lb in h5. 
 assert (h6: lb (Im (Full_set T) q) (r * (-p))).
   unfold ub. intros b h7.
   unfold In in h7. inversion h7.
   assert (h8: b = (p * (-p)) + b). 
     rewrite comp_prod. rewrite comm_sum. rewrite zero_sum.
     reflexivity.
   assert (h9: (p * (-p))+b = (p+b) * ((-p)+b)).
     apply dist_prod_r.
   assert (h10: b = (p+b) * ((-p)+b)).
     congruence.
   assert (h11:le r (p+b)).
     apply h5.
     unfold In.
     rewrite H0.
     apply Im_intro with (x := x).
     apply H.
     unfold pq.
     unfold compose. reflexivity.
   assert (h12:le (-p) (-p+b)).
     unfold le.  
     assert (h13: (-p)+((-p)+b) = ((-p)+(-p))+b).  
       rewrite assoc_sum.  reflexivity. 
     assert (h14: (-p)+(-p) = (-p)).      
       apply idem_sum.
     rewrite h14 in h13.
     apply h13.
   assert (h15 : le (r * -p) ((p+b) * (-p+b))).
     apply mono_prod.
   assumption.  assumption.
   rewrite  h10. assumption.
assert (h11: le (r * (-p)) q').
  apply h1. assumption.
assert (h12: le p p).
  apply refl_le.
assert (h13: le (p+(r*-p)) (p+q')).
  apply mono_sum. assumption. assumption.
assert (h14: p+(r * -p) = (p+r) *( p+-p)).
  apply dist_prod.
assert (h15: (p+r) * (p+-p) = p+r).  
  rewrite comp_sum.  rewrite one_prod.  reflexivity. 
assert (h16: le r (p+r)). unfold le.  
    assert (h18: r+r = r).  apply idem_sum.
    assert (h19: r + (p + r) = (r + r) + p). 
      rewrite (comm_sum _ p r).
      apply assoc_sum. 
    rewrite h19. rewrite h18. 
      rewrite (comm_sum _ p r). reflexivity. 
rewrite <- h15 in h16.
rewrite <- h14 in h16.
apply trans_le with (x := r) (y := (p+(r*-p))) (z:=p+q').
assumption. assumption.
unfold Inf. unfold InfGen. unfold inf. split.
assumption. assumption.
Qed.

Variable Jt:Type.
Definition Sup' {T:Type} (f:T->Bt) (b:Bt) := 
  SupGen T (Full_set T) f b.

Lemma sup_le : forall {T:Type} (f:T->Bt) (b:Bt), 
  Sup' f b -> forall (j:T), le (f j) b.
intros T f b h1 j.
    unfold Sup' in h1. red in h1. red in h1. destruct h1 as [h1 ?].
    red in h1. 
    apply (h1 (f j)). 
    pose proof (Full_intro _ j) as h2.
    apply Im_intro with j. assumption. reflexivity.
Qed.


End infinite.
Arguments Bcompi [B] _ _ _.
Arguments SupGen [B] _ _ _ _.
Arguments InfGen [B] _ _ _ _.
Arguments Sup [B] _ _ _.
Arguments Inf [B] _ _ _.
Arguments comp_ind [B] _ _ _.
Arguments infnt_de_mor1 [B] _ _ _ _.
Arguments infnt_de_mor2 [B] _ _ _ _.
Arguments infnt_de_mor3 [B] _ _ _ _.
Arguments infnt_de_mor4 [B] _ _ _ _.
Arguments complete [B] _.
Arguments complete_sup [B] _ _ _.
Arguments complete_inf [B] _ _ _.
Arguments infnt_assoc_sup [B] _ _ _ _ _ _ _ _ _.
Arguments infnt_assoc_inf [B] _ _ _ _ _ _ _ _ _.
Arguments infnt_distr_1_sup [B] _ _ _ _ _.
Arguments infnt_distr_1_inf [B] _ _ _ _ _.
Arguments Sup' [B] _ _ _.
Arguments sup_le [B] _ _ _ _ _.

