(* Copyright (C) 2014-2015, Daniel Wyckoff, except for the portions so labeleled as (*Schepler*) which I got from Daniel Schepler.  
My functions will be labeled (*Wyckoff*)*)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)


Require Export EnsemblesImplicit.
Require Export ImageImplicit.
Require Import Classical.
Require Import Image.
Require Import LogicUtilities.
Require Import Setoid.
Require Import FunctionProperties.
Require Import Arith.
Require Import Equality.
Require Import Infinite_sets.


(*Wyckoff*)
Lemma in_add_eq : forall {T:Type} (A:Ensemble T) (x:T),
                    In A x -> Add A x = A.
intros T A x h1.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros y h2.
destruct h2 as [y h2l | y h2r]. assumption.
destruct h2r; subst. assumption.
(* >= *)
red. intros y h2.
left. assumption.
Qed.

(*Wyckoff*)
Lemma add_not_empty : forall {T:Type} (A:Ensemble T) (x:T),
                        Add A x <> Empty_set _.
intros T A x. intro h1.
pose proof (Add_intro2 _ A x) as h2.
rewrite h1 in h2.
contradiction.
Qed.

(*Wyckoff*)
Lemma add_empty_sing : forall {T:Type} (x:T), 
                         Add (Empty_set _) x =
                         Singleton x.
intros T x.
apply Extensionality_Ensembles.
red. split.
red.
intros y h1.
destruct h1; try contradiction; auto.
red.
intros y h2. right; auto.
Qed.

(*Wyckoff*)
Lemma sub_add_comm : 
  forall {T:Type} (A:Ensemble T) 
         (x y:T),  (x <> y) ->
    (Subtract (Add A x) y) = (Add (Subtract A y) x).
intros T A x y h0.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros z h1.
destruct h1 as [h1 h2].
destruct h1 as [h1 | z h3].
left. constructor; auto.
right; auto.
(* >= *)
red. intros z h1.
destruct h1 as [z h1 | z h2].
destruct h1 as [h1 h3].
constructor. constructor; auto. auto.
destruct h2; subst.
constructor. apply Add_intro2.
intro h1. destruct h1. 
contradict h0. auto.
Qed.

(*Wyckoff*)
Lemma sub_add_compat_nin : 
  forall {T:Type} (A:Ensemble T) (x:T),
    ~ In A x ->
    (Subtract (Add A x) x) = A.
intros T A x h1.
apply Extensionality_Ensembles.
red. split.
(* => *)
red.
intros y h2.
destruct h2 as [h2 ?].
destruct h2; try contradiction; auto.
(* <= *)
red.
intros y h2.
constructor. left; auto.
intro h3.
destruct h3; subst.
contradiction.
Qed.                          

(*Wyckoff*)
Lemma sub_add_compat_in :
  forall {T:Type} (A:Ensemble T) (x:T),
    In A x ->
    (Subtract (Add A x) x) = Subtract A x.
intros T A x h1.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros y h2.
destruct h2 as [h3 h4].
destruct h3 as [y h3 | y h5].
constructor; auto. contradiction.
(* >= *)
red.
intros y h2.
destruct h2. 
constructor; try constructor; auto.
Qed.

(*Wyckoff*)
Lemma add_nin_sub_compat :
  forall {T:Type} (A B:Ensemble T) (x:T),
    ~Ensembles.In A x ->
    Add A x = Add B x ->
    A = Subtract B x.
intros T A B x h1 h2.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros y h3.
pose proof (Add_intro1 _ A x y h3) as h4.
rewrite h2 in h4.
pose proof (Add_inv _ B x y h4) as h5.
destruct h5 as [h5l | h5r].
constructor; auto.
intro h6.
destruct h6; subst; contradiction.
subst. contradiction.
(* >= *)
red. 
intros y h3.
destruct h3 as [h3 h4].
assert (h5: x <> y).
  intro. subst.
  pose proof (In_singleton _ y). contradiction.
clear h4.
pose proof (Add_intro1 _ B x y h3) as h4.
rewrite <- h2 in h4.
destruct h4 as [y h4l | y h4r]. auto.
destruct h4r; subst. 
contradict h5. auto.
Qed.

(*Wyckoff*)
Lemma add2 : forall {T:Type} (A:Ensemble T) (x:T), 
               Add (Add A x) x = Add A x.
intros T A x.
unfold Add.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros y h1.
destruct h1 as [y h2 | y h3].
auto.
destruct h3.
apply Union_intror. constructor.
(* >= *)
red. intros y h1.
constructor. auto.
Qed.

(*Wyckoff*)
Lemma add_comml : forall {T:Type} (A:Ensemble T) (x y:T),
                   Included (Add (Add A x) y) (Add (Add A y) x).
intros T A x y.
red. 
intros z h1.
destruct h1 as [z h2|z h3].
destruct h2 as [z h4 | z h5].
left. left. auto.
destruct h5; subst.
right. constructor.
destruct h3; subst.
left. right. constructor.
Qed.

(*Wyckoff*)
Lemma add_comm : forall {T:Type} (A:Ensemble T) (x y:T),
                   Add (Add A x) y = Add (Add A y) x.
intros;
apply Extensionality_Ensembles; red; split;
apply add_comml.
Qed.

(*Wyckoff*)
Lemma union_add_comm : forall {T:Type} (A B:Ensemble T) (x:T),
                    Union (Add A x) B = Add (Union A B) x.
intros T A B x.
apply Extensionality_Ensembles.
red. split.
(* <= *) red.
intros y h1.
destruct h1 as [y h1 | y h2].
destruct h1 as [y h1 | y h3].
left. left. assumption.
destruct h3. subst. right. constructor.
left. right. assumption.
(* >= *)
red. intros y h1.
destruct h1 as [y h1 | y h2].
destruct h1 as [y h1 | y h3].
left. left. assumption.
right. assumption.
destruct h2; subst.
left. right. constructor.
Qed.

(*Wyckoff*)
Lemma couple_add_sing : forall {T:Type} (x y:T),
                         Couple x y = Add (Singleton x) y.
intros T x y.
apply Extensionality_Ensembles.
red. split. 
red.
intros z h1.
destruct h1. left. constructor.
right. constructor.
red.
intros z h1.
destruct h1 as [z h1l | z h1r].
destruct h1l; subst. left.
destruct h1r. right.
Qed.

(*Wyckoff*)
Lemma couple_add_add : forall {T:Type} (x y:T),
                         Couple x y = Add (Add (Empty_set _) x) y.
intros T x y.
rewrite couple_add_sing.
rewrite add_empty_sing.
reflexivity.
Qed.

(*Wyckoff*)
Lemma couple_singleton : forall {T:Type} (x:T), 
                           Couple x x = Singleton x.
intros T x.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros ? h1.
destruct h1; constructor.
(* >= *)
red. intros ? h1.
destruct h1; constructor.
Qed.

(*Wyckoff*)
Lemma empty_union : forall {T:Type} (S:Ensemble T),
                      Union (Empty_set T) S = S.
intros T S.
apply Extensionality_Ensembles.
red. split.
  (* <= *)
  red.
  intros x h1.
  destruct h1; [contradiction | assumption].
  (* >= *)
  red.
  intros x h1.
  apply Union_intror; assumption.
Qed.

(*Wyckoff*)
Lemma empty_intersection : forall {T:Type} (S:Ensemble T),
                             Intersection (Empty_set T) S =
                             Empty_set T.
  intros T S.
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red.
  intros x h1.
  destruct h1; contradiction.
  (* >= *)
  auto with sets.
Qed.

(*Wyckoff*)
Lemma singleton_inj : forall {T:Type} (a b:T), 
                        Singleton a = Singleton b ->
                        a = b.
intros T a b h1.
pose proof (In_singleton _ a) as h2.
rewrite h1 in h2.
destruct h2; reflexivity. 
Qed.

(*Wyckoff*)
Lemma im_singleton : forall {T U:Type} (x:T) (f:T->U),
                       Im (Singleton x) f = Singleton (f x).
intros T U x f.
apply Extensionality_Ensembles.
red. split.
red.
intros y h1.
destruct h1 as [z h2]. subst.
destruct h2; subst.
constructor; auto.
red. 
intros y h1.
destruct h1; subst.
apply Im_intro with x; auto; constructor.
Qed.

(*Wyckoff*)
Lemma finite_cardinal_unq : 
  forall {T:Type} (S:Ensemble T), 
    Finite  S -> exists! n:nat, cardinal _ S n.
intros T S h1.
pose proof (finite_cardinal _ _ h1) as h2.
destruct h2 as [n h2].
exists n.
red.
split; try assumption.
apply cardinal_unicity; assumption.
Qed.
           
(*Wyckoff*)                       
Lemma singleton_inc : forall {T:Type} (a:T) (A:Ensemble T),
                        Included A (Singleton a) ->
                        A = Empty_set _ \/ A = Singleton a.
intros T a A h0.
pose proof (Singleton_is_finite _ a) as h6.
assert (h7:cardinal _ (Singleton a) 1).
assert (h7:Singleton a = Add (Empty_set _) a).
unfold Add.
rewrite empty_union. reflexivity.
rewrite h7.
constructor. constructor. intro. contradiction.
pose proof (Finite_downward_closed _ _ h6 _ h0) as h8.
pose proof (finite_cardinal _ _ h8) as h9.
destruct h9 as [n h9].
pose proof (incl_card_le _ _ _ _ _ h9 h7 h0) as h10.
pose proof (le_pred n 1 h10) as h11.
simpl in h11.
pose proof (le_n_0_eq _ h11) as h12.
unfold pred in h12.

destruct (zerop n) as [? | h13].
subst.        
left.
apply cardinalO_empty. assumption.
assert (h14: n = S O). auto with arith.
subst.
right.
apply Extensionality_Ensembles.
red. split. assumption.
red.
intros x h14.
destruct h9.    
pose proof (lt_irrefl 0).
contradiction.        
assert (h15:a = x0).
unfold Included in h0.
pose proof (Add_intro2 _ A x0) as h15.
specialize (h0 _ h15).
destruct h0; auto.
subst.
destruct h14; subst.
apply Add_intro2.
Qed.

(*Wyckoff*)
Lemma singleton_inc_int : 
  forall {T:Type} (a:T) (A:Ensemble T),
    Intersection A (Singleton a) =  Empty_set _ \/ 
    Intersection A (Singleton a) = Singleton a.
intros T a A.
assert (h1:Included (Intersection A (Singleton a)) (Singleton a)). auto with sets.
apply singleton_inc.
assumption.
Qed.

(*Wyckoff*)
Lemma inclusion_intersection : forall {T:Type} (P S R : Ensemble T), 
  (Included P S /\ Included P R) <-> Included P (Intersection S R).

intros T P S R.
split.
(*left*)
intro h1.
destruct h1 as [h2 h3].
red.
intros x h4.
constructor; auto with sets.
(*right*)
intro h1.
split; red.
  (*left*)
  intros x h5.
  red in h1.
  pose proof (h1 x h5) as h6.
  inversion h6; trivial.
  (*right*)
  intros x h7.
  pose proof (h1 x h7) as h6. 
  inversion h6; trivial.
Qed.

(*Wyckoff*)
Lemma intersection_preserves_inclusion : forall {T:Type} (P Q R:Ensemble T), 
  Included P Q -> Included (Intersection R P) (Intersection R Q).
intros T P Q R h1.
red.
intros x h2.
destruct h2 as [x h2l h2r].
auto with sets.
Qed.

(*Wyckoff*)
Lemma inclusion_iff_intersection_eq : forall {T:Type} (P Q:Ensemble T), 
  Included P Q <-> Intersection P Q = P.
intros T P Q.
split.
(* -> *)
intro.
apply Extensionality_Ensembles.
red. split; auto with sets.
(* <= *)
intro h1.
red.
intros x h2.
rewrite <- h1 in h2.
destruct h2; assumption.
Qed.

(*Wyckoff*)
Lemma inclusion_iff_union : forall {T:Type} (P Q:Ensemble T), 
  Included P Q <->  Union P Q = Q.
intros T P Q.
split. intro h1.
apply Extensionality_Ensembles.
red. split.
red. intros x h2. destruct h2; auto with sets.
red. intros; right; auto.
intro h1. red. intros.  rewrite <- h1. auto with sets.
Qed.


(*Wyckoff*)
Lemma empty_inclusion : forall {T:Type} (S:Ensemble T), Included (Empty_set _) S.
intros T S.
red. intros; contradiction.
Qed.

(*Wyckoff*)
Lemma full_inclusion : forall {T:Type} (S:Ensemble T), 
                         Included S (Full_set _).
intros; red; intros; constructor.
Qed.                                                                

(*Wyckoff*)
Lemma inclusion_reflexive : forall {T:Type} (S:Ensemble T),
                    Included S S.
auto with sets.
Qed.

(*Wyckoff*)
Lemma intersection_full_set : forall {T:Type} (S:Ensemble T), 
  Intersection S (Full_set T) = S.
intros T S.
apply Extensionality_Ensembles.
red. split.
  (* <= *) red. intros ? h. destruct h; assumption.
  (* >= *) red. intros x ?. constructor. assumption. constructor.
Qed.

(*Wyckoff*)
Lemma included2_impl_union_included : forall {T:Type} (A B C:Ensemble T), 
  ((Included A C) /\ (Included B C)) <-> Included (Union A B) C.
intros T A B C.
split.
(* -> *)
unfold Included.
intro h1.
destruct h1 as [h1 h2].
intros x h3.
apply Union_inv in h3.
destruct h3; auto with sets.
(* <- *)
intro h1.
split.
assert (Included A (Union A B)). auto with sets.
auto with sets.
assert (Included B (Union A B)). auto with sets.
auto with sets.
Qed.


Section complement.

(*Wyckoff*)
Lemma complement_inclusion: forall {Y:Type} (S T:Ensemble Y),
  Included S T -> Included (Ensembles.Complement T) (Ensembles.Complement S).
Proof.
intros.
red; intros.
red; intro.
contradiction H0.
auto with sets.
Qed.

(*Wyckoff*)
Lemma excl_middle_empty : (forall {T:Type} (S:(Ensemble T)), Intersection 
  S (Ensembles.Complement S) = Empty_set _).
intros T S.
apply Extensionality_Ensembles. red. split.
(*left*)
red.
intros x h1.
inversion h1; contradiction.
(*right*)
red.
intros; contradiction.
Qed.

(*Wyckoff*)
Lemma excl_middle_full : forall {T:Type} (S:Ensemble T), Union
  S (Ensembles.Complement S) = Full_set _.
intros T S.
apply Extensionality_Ensembles. red. split.
(*left*)
red. intros. constructor. auto with sets.
(*right*)
red. intros x h1.
case (classic (In S x)) as [h2 | h3].
  apply Union_introl; trivial.
  apply Union_intror.
  auto with sets.
Qed.

(*Wyckoff*)
Lemma complement_full : forall T:Type, Ensembles.Complement (Full_set T) = Empty_set T.
intro T.
apply Extensionality_Ensembles. red. unfold Included. split.
(* left *)
intros x h2.
unfold Ensembles.Complement in h2. red in h2.
elim h2. apply Full_intro.
(* right *)
intros x h3. contradiction.
Qed.

(*Wyckoff*)
Lemma complement_empty : forall T:Type, Ensembles.Complement (Empty_set T) = Full_set T.
intro T.
pose proof (Complement_Complement _ (Full_set T)) as h1.
rewrite complement_full in h1.
assumption.
Qed.

(*Wyckoff*)
Lemma complement_meets_non_subset : forall {T:Type} (P Q : Ensemble T), 
  (Inhabited P) ->
  (~Included P Q <-> Inhabited (Intersection (Ensembles.Complement Q) P)).
intros T P Q h1.
split.
(*left*)
intro h2. 
unfold Included in h2.
pose proof (not_all_ex_not _ _ h2) as h4.
elim h4. intros x h5.
assert (h6: (In P x) /\ (~In Q x)).
  tauto.
destruct h6 as [h7 h8].
assert (h9: In (Intersection (Ensembles.Complement Q) P) x).
  constructor; auto with sets. 
apply Inhabited_intro with x; trivial.
(*right*)
intro h2.
unfold Included.
inversion h2 as [x h3].
assert (h4: exists x:T, ~(In P x -> In Q x)).
  exists x.
  assert (h5: (In P x) /\ (~ In Q x)).
    split.
    inversion h3 as [h6 h7]. assumption.
    inversion h3 as [y h6 h7].
    compute in h7. auto.
  tauto.
apply ex_not_not_all; trivial.
Qed.


(*Wyckoff*)
Lemma included_empty_complement_int : forall {T:Type} (A B:Ensemble T), 
  Included A B <-> Intersection A (Ensembles.Complement B) = Empty_set _.
intros T A B.
split.
(* -> *)
intros h1.
apply Extensionality_Ensembles. red.
  split.
  (* <= *)
  red in h1.
  red. intros x h2.
  destruct h2 as [x h2l h2r].
  specialize (h1 _ h2l).
  assert (h3:In (Intersection B (Ensembles.Complement B)) x).
    auto with sets.
  rewrite excl_middle_empty in h3. assumption.
  (* >= *) auto with sets.
(* <- *)
intro h1.
red. 
intros x h2.
apply NNPP.
intro h3.
assert (h4:In (Intersection A (Ensembles.Complement B)) x).
  constructor; assumption.
rewrite h1 in h4.
contradiction.
Qed.

(*Wyckoff*)
Lemma setminus_inc : forall {T:Type} (A B:Ensemble T),
                       Included (Setminus A B) A.
intros T A B. red.
intros x h1.
destruct h1; auto with sets.
Qed.


(*Wyckoff*)
Lemma setminus_int_complement : forall {T:Type} (A B:Ensemble T), Setminus A B = 
  Intersection A (Ensembles.Complement B).
intros T A B.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros ? h1.
destruct h1 as [h2 h3].
constructor; auto with sets. 
(* >= *)
red. intros x h1.
destruct h1 as [h2 h3].
constructor; auto with sets.
Qed.

(*Wyckoff*)
Lemma setminus_sub_sup : forall {T:Type} (A B:Ensemble T), Included A B -> 
  Setminus A B = Empty_set _.
intros T A B h1.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros x h2.
destruct h2 as [h3 h4].
red in h1.
apply h1 in h3.
contradiction.
(* >= *) auto with sets.
Qed.

(*Wyckoff*)
Lemma int_setminus : forall {T:Type} (A X:Ensemble T),
                       Intersection A (Setminus X A) = Empty_set _.
intros T A X.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros x h1.
destruct h1 as [x h2 h3].
destruct h3. contradiction.
(* >= *)
auto with sets.
Qed.

(*Wyckoff*)
Lemma setminus_same_int : forall {T:Type} (X Y:Ensemble T),
                            Setminus X (Intersection X Y) =
                            Setminus X Y.
intros T X Y.
apply Extensionality_Ensembles.
red. split.
red.
intros x h1.
destruct h1 as [h1 h2].
constructor. assumption. auto with sets.
red.
intros x h1. constructor.
destruct h1 as [h1 h2].
assumption.
intro h2.
destruct h2; destruct h1; contradiction.
Qed.

(*Wyckoff*)
Lemma decompose_int_setminus : 
  forall {T:Type} (X Y:Ensemble T),
    X = Union (Intersection X Y) (Setminus X Y).
intros T X Y.
apply Extensionality_Ensembles.
red. split.
red.
intros x h1.
destruct (classic (In Y x)).
left. constructor; auto.
right. constructor; auto.
red.
intros x h1.
destruct h1 as [x h1l | x h1r].
destruct h1l; auto.
destruct h1r; auto.
Qed.

(*Wyckoff*)
Lemma setminus_add1 : forall {T:Type} (E:Ensemble T) (x:T),
                        ~ In E x ->
                        Setminus (Add E x) E = Singleton x.
intros T E x h1.
apply Extensionality_Ensembles.
red. split.
red.
intros y h2.
destruct h2 as [h2 h3].
destruct h2 as [y h2l | y h2r]. contradiction.
assumption.
red. intros y h2. destruct h2; subst. constructor; auto.
right; auto. constructor.
Qed.

   
(*Wyckoff*)
Lemma add_subtract_a : 
  forall {T:Type} (A:Ensemble T) (a:T),
    Subtract (Add A a) a = Subtract A a.
intros T A a.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros x h1.
destruct h1 as [h1 h2].
destruct h1 as [x h1a | x h1b].
constructor; auto. contradiction.
(* >= *)
red.
intros x h1.
destruct h1 as [h1 h2].
pose proof (Add_intro1 _ A a x h1).
constructor; auto.
Qed.

(*Wyckoff*)
Lemma setminus_couple_eq :
  forall {T:Type} (A:Ensemble T) (a a':T), 
    Setminus A (Couple a a') = Subtract (Subtract A a) a'.
intros T A a a'.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros x h1.
destruct h1 as [h1 h2].
constructor. constructor. assumption.
intro h3. destruct h3; subst.
pose proof (Couple_l _ a a'); contradiction.
intro h3. destruct h3; subst.
pose proof (Couple_r _ a a'); contradiction.
(* >= *)
red. intros x h1.
destruct h1 as [h1 h2].
destruct h1 as [h1a h1b].
constructor. assumption.
intro h3.
destruct h3.
pose proof (In_singleton _ a). contradiction.
pose proof (In_singleton _ a'). contradiction.
Qed.

(*Wyckoff*)
Lemma subtract2_comm : 
  forall {T:Type} (S:Ensemble T) (a b:T),
    Subtract (Subtract S a) b =
    Subtract (Subtract S b) a.
  intros T S a b.
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red. intros x h1.
  destruct h1 as [h1a h1b].
  destruct h1a as [h1aa h1ab].
  constructor. constructor. 
  assumption. assumption. assumption.
  (* >= *)
  red. intros x h1.
  destruct h1 as [h1a h1b].
  destruct h1a as [h1aa h1ab].
  constructor. constructor.
  assumption. assumption. assumption.
Qed.

(*Wyckoff*)
Lemma couple_comm : 
  forall {T:Type} (a b:T), 
    Couple a b = Couple b a.
  intros T a b.
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red. intros ? h1.
  destruct h1; subst; constructor.
  (* >= *)
  red. intros ? h1.
  destruct h1; subst; constructor.
Qed.

(*Wyckoff*)
Lemma union_singleton_eq_setminus :
  forall {T:Type} (A B:Ensemble T) (a b:T), 
    Add A a = Add B b -> 
    Setminus A (Couple a b) = Setminus B (Couple a b).
intros T A B a b h1.
assert (h2:Setminus (Add A a) (Couple a b) = 
           Setminus (Add B b) (Couple a b)).
  f_equal; auto.
repeat rewrite setminus_couple_eq in h2.
rewrite add_subtract_a in h2.
symmetry in h2.
rewrite subtract2_comm in h2.
rewrite add_subtract_a in h2.
repeat rewrite <- setminus_couple_eq in h2.
symmetry.
rewrite couple_comm in h2.
assumption.
Qed.

End complement.

Section family1.
Variable It T:Type.
Definition Family := Ensemble (Ensemble T).

(*Schepler*)
(*Copied from Zorn's Lemma Coq user contribution *)
Inductive FamilyUnion (F:Family): Ensemble T :=
  | family_union_intro: forall (S:Ensemble T) (x:T),
    In F S -> In S x -> In (FamilyUnion F) x.

(*Schepler*)
Inductive FamilyIntersection (F:Family) : Ensemble T :=
  | family_intersection_intro: forall x:T,
    (forall S:Ensemble T, In F S -> In S x) ->
    In (FamilyIntersection F) x.

(*Schepler*)
Definition IndexedFamily := It -> Ensemble T.
Variable FI:IndexedFamily.

(*Schepler*)
Inductive IndexedUnion : Ensemble T :=
  | indexed_union_intro: forall (i:It) (x:T),
    In (FI i) x -> In IndexedUnion x.

(*Schepler*)
Inductive IndexedIntersection : Ensemble T :=
  | indexed_intersection_intro: forall (x:T),
    (forall i:It, In (FI i) x) -> In IndexedIntersection x.

(*Schepler*)
Definition ImageFamily : Family:=
  Im (Full_set _) FI.

(*Schepler*)
Lemma indexed_to_family_union: IndexedUnion = FamilyUnion (ImageFamily).
Proof.
apply Extensionality_Ensembles.
unfold Same_set.
unfold Included.
intuition.
destruct H.
apply family_union_intro with (FI i).
apply Im_intro with i.
constructor.
reflexivity.
assumption.

destruct H.
destruct H.
apply indexed_union_intro with x0.
rewrite <- H1.
assumption.
Qed.                                                        

(*Schepler*)
Lemma indexed_to_family_intersection:
  IndexedIntersection = FamilyIntersection ImageFamily.
Proof.
apply Extensionality_Ensembles.
unfold Same_set.
unfold Included.
intuition.
constructor.
intros.
destruct H.
destruct H0.
rewrite H1.
apply H.

constructor.
intro.
destruct H.
apply H.
apply Im_intro with i.
constructor.
reflexivity.
Qed.
End family1.

(*Wyckoff*)
Lemma family_union_add : forall {T:Type} (F:Family T) (S:Ensemble T),
  FamilyUnion _ (Add F S) = Union S (FamilyUnion _ F).
intros T F S.
apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red. intros x h4.
  unfold Add in h4.
  inversion h4 as [A y h5 h6].
  pose proof (Union_inv _ _ _ A h5) as h7.
  case h7 as [h8 | h9].
  (*h8*)
  apply Union_intror. 
  apply family_union_intro with A; assumption.
  (*h9*)
  left. 
  pose proof (Singleton_inv _ _ _ h9) as h12.
  rewrite <- h12 in h6; assumption.
  (* >= *)
  red. intros x h2.
  pose proof (Union_inv _ _ _ x h2) as h3.
  case h3 as [h4 | h5].
  (*h4*)
  pose proof (Add_intro2 _ F S) as h5.
  apply family_union_intro with S; assumption.
  (*h5*)
  inversion h5 as [A y h6 h7]. 
  pose proof (Add_intro1 _ F S) as h8.
  pose proof (h8 A h6) as h9.
  apply family_union_intro with A; assumption.
Qed.

Section family2.

(*Schepler*)
Lemma empty_indexed_intersection: forall {T:Type}
  (F:IndexedFamily False T),
  IndexedIntersection _ _ F = (Full_set _).
Proof.
intros.
apply Extensionality_Ensembles; red; split; red; intros;
  auto with sets.
constructor.
constructor.
destruct i.
Qed.

Variable T:Type.
(*Schepler*)
Lemma empty_family_union: FamilyUnion T (Empty_set _) =
                          (Empty_set _).
Proof.
apply Extensionality_Ensembles.
unfold Same_set.
unfold Included.
intuition.
destruct H.
contradiction H.
contradiction H.
Qed.

(*Wyckoff*)
Definition chain (F:Family T) : Prop := 
  forall (S1 S2:Ensemble T), In F S1 -> 
    In F S2 -> Included S1 S2 \/ Included S2 S1.

End family2.

Arguments FamilyUnion [T] _ _.
Arguments FamilyIntersection [T] _ _.
Arguments IndexedUnion _ [T] _ _.
Arguments IndexedIntersection _ [T] _ _.

Section Power_set_Algebra_axioms.
Variable Xt:Type.
(*Wyckoff*)
Lemma assoc_sum_psa : forall N M P:(Ensemble Xt), (Union N (Union M P)) = 
  (Union (Union N M) P).
intros N M P.
apply Extensionality_Ensembles.  unfold Same_set.
split; unfold Included.
(*left*)
intros x H.
inversion H.
assert (h1 : In (Union N M) x). 
apply Union_introl. assumption.
apply Union_introl. assumption.
inversion H0.
assert (h1 : In (Union N M) x). 
apply Union_intror. assumption.
apply Union_introl. assumption.
apply Union_intror. assumption.
(*right*)
intros x H.
inversion H. inversion H0.
apply (Union_introl _ _ _ _ H2).
assert (h2: In (Union M P) x).
apply (Union_introl _ _ _ _ H2).
apply (Union_intror _ _ _ _ h2).
assert (h2: In (Union M P) x).
apply (Union_intror _ _ _ _ H0).
apply (Union_intror _ _ _ _ h2).
Qed.

(*Wyckoff*)
Lemma assoc_prod_psa : forall N M P:(Ensemble Xt), 
  (Intersection N (Intersection M P)) = 
  (Intersection (Intersection N M) P).
intros N M P.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 h4].
inversion h3 as [x'' h5 h6 h7].
assert (h8: In (Intersection N M) x).
apply Intersection_intro; assumption.
apply Intersection_intro; assumption.
(*Learn how to automate the above duplication*)
(*right*)
intros x h9.
inversion h9 as [x' h10 h11 h12].
inversion h10 as [x'' h13 h14 h15].
assert (h16: In (Intersection M P) x).
apply Intersection_intro; assumption.
apply Intersection_intro; assumption.
Qed.

(*Wyckoff*)
Lemma comm_sum_psa : forall N M:(Ensemble Xt), Union N M = Union M N.
intros N M.
apply Extensionality_Ensembles.
unfold Same_set. 
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 | x'' h4 h5].
apply Union_intror; assumption.
apply Union_introl; assumption.
(*right*)
intros x h6.
inversion h6 as [x' h7 h8 | x'' h9 h10].
apply Union_intror; assumption.
apply Union_introl; assumption.
Qed.

(*Wyckoff*)
Lemma comm_prod_psa : 
  forall N M:(Ensemble Xt), Intersection N M = Intersection M N.
intros N M.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 h4].
apply Intersection_intro; assumption.
(*right*)
intros x h5.
inversion h5 as [x' h6 h7 h8].
apply Intersection_intro; assumption.
Qed.

(*Wyckoff*)
Lemma abs_sum_psa  : 
  forall N M:(Ensemble Xt), Union N (Intersection N M) = N.
intros N M.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 | x'' h4 h5].
assumption.
inversion h4 as [x''' h6 h7].
assumption.
(*right*)
intros x h8.
apply Union_introl. assumption.
Qed.

(*Wyckoff*)
Lemma abs_prod_psa  : 
  forall N M:(Ensemble Xt), Intersection N (Union N M) = N.
intros N M.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 h4].
assumption.
(*right*)
intros x h5.
assert (h6:In (Union N M) x).
apply Union_introl; assumption.
apply Intersection_intro; assumption.
Qed.

(*Wyckoff*)
Lemma dist_sum_psa :
  forall N M P:(Ensemble Xt), 
    Intersection P (Union N M) = 
    Union (Intersection P N) (Intersection P M).
intros N M P.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 h4].
inversion h3 as [x'' h5 h6 | x'' h7 h8].
assert (h9: In (Intersection P N) x).
apply Intersection_intro; assumption.
apply Union_introl; assumption.
assert (h10: In (Intersection P M) x).
apply Intersection_intro; assumption.
apply Union_intror; assumption.
(*right*)
intros x h11.
inversion h11 as [x' h12 h13 | x'' h14 h15].
inversion h12 as [x''' h16 h17 h18].
assert (h19 : In (Union N M) x).
apply Union_introl; assumption.
apply Intersection_intro; assumption.
inversion h14 as [x''' h20 h21 h22].
assert (h23: In (Union N M) x).
apply Union_intror; assumption.
apply Intersection_intro; assumption.
Qed.

(*Wyckoff*)
Lemma dist_prod_psa : 
  forall N M P:(Ensemble Xt), 
    Union P (Intersection N M) = 
    Intersection (Union P N) (Union P M).
intros N M P.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 | x'' h4 h5].
assert (h6: In (Union P N) x). auto with sets.
assert (h7: In (Union P M) x). auto with sets.
auto with sets.
inversion h4.
assert (h8: In (Union P N) x). auto with sets.
assert (h9: In (Union P N) x). auto with sets.
auto with sets.
(*right*)
intros x h10.
inversion h10 as [x' h11 h12 h13].
inversion h11. 
auto with sets.
inversion h12.
auto with sets.
assert (h14: In (Intersection N M) x). 
auto with sets. auto with sets.
Qed.

(*Wyckoff*)
(*Uses Law of Excluded Middle*)
Lemma comp_sum_psa: 
  forall N:(Ensemble Xt), Union N (Ensembles.Complement N) = Full_set Xt.
intros N.
apply Extensionality_Ensembles.
unfold Same_set. 
split; unfold Included.
(*left*)
intros x h1.
apply Full_intro.
(*right*)
intros x h2.
assert (h3: ~(In N x) \/ (In N x)). 
    tauto.
elim h3; auto with sets.
Qed.

(*Wyckoff*)
Lemma comp_prod_psa:
  forall N:(Ensemble Xt), Intersection N (Ensembles.Complement N) = Empty_set Xt.
intros N.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
contradict h1.
intro h2.
inversion h2 as [x' h3 h4].
unfold In in h3.  unfold In in h4. unfold Ensembles.Complement in h4.
tauto.
(*right*)
intros x h5.
inversion h5.
Qed.
End Power_set_Algebra_axioms.

(*Wyckoff*)
Lemma decompose_setminus_inc : 
  forall {T:Type} (X Y:Ensemble T),
    Included Y X ->
    X = Union Y (Setminus X Y).
intros T X Y h1.
rewrite (decompose_int_setminus X Y) at 1.
rewrite inclusion_iff_intersection_eq in h1.
rewrite comm_prod_psa in h1.
rewrite h1.
reflexivity.
Qed.

(*Wyckoff*)
Lemma union_eq_empty : forall {T:Type} (A B:Ensemble T),
  Union A B = Empty_set _ -> A = Empty_set _ /\ B = Empty_set _.
intros T A B h1.
split; apply Extensionality_Ensembles; split; auto with sets.
red. intros x h2. pose proof (Union_introl _ A B _ h2) as h3.
rewrite h1 in h3. assumption.
red. intros x h2. pose proof (Union_introl _ B A _ h2) as h3.
rewrite comm_sum_psa in h3.
rewrite h1 in h3. assumption.
Qed.

(*Wyckoff*)
Lemma intersection_eq_full : forall {T:Type} (A B:Ensemble T),
  Intersection A B = Full_set _ -> A = Full_set _ /\ B = Full_set _. 
intros T A B h1. split; apply Extensionality_Ensembles; split;
try (red; intros; constructor).
red. intros x h2.
rewrite <- h1 in h2. destruct h2; assumption.
red. intros x h2.
rewrite <- h1 in h2. destruct h2; assumption.
Qed.

Section Symdiff.

(*Wyckoff*)
Definition Symdiff {T:Type} (A B:Ensemble T) :=
  Union (Setminus A B) (Setminus B A).

(*Wyckoff*)
Lemma symdiff_ref : forall {T:Type} (A:Ensemble T),
                        Symdiff A A = Empty_set _.
intros T A.
unfold Symdiff.
rewrite setminus_sub_sup; auto with sets.
Qed.

(*Wyckoff*)
Lemma symdiff_comm : forall {T:Type} (A B:Ensemble T),
                       Symdiff A B = Symdiff B A.
intros T A B. unfold Symdiff. auto with sets.
Qed.

(*Wyckoff*)
Lemma symdiff_empty_iff_eq : 
  forall {T:Type} (A B:Ensemble T),
    Symdiff A B = Empty_set _ <-> A = B.
intros T A B.
split.
(* -> *)
intro h1.
unfold Symdiff in h1.
apply Extensionality_Ensembles.
red. split.
(* <= *) red.
intros x h2.
apply NNPP.
intro h3.
pose proof (Setminus_intro _ A B x h2 h3) as h4.
pose proof (Union_introl _ (Setminus A B) (Setminus B A) x h4) as h5.
rewrite h1 in h5. contradiction.
(* >= *)
intros x h2.
apply NNPP.
intro h3.
pose proof (Setminus_intro _ B A x h2 h3) as h4.
pose proof (Union_introl _ (Setminus B A) (Setminus A B) x h4) as h5.
rewrite comm_sum_psa in h5.
rewrite h1 in h5. contradiction.
(* <- *)
intro h1.
rewrite h1.
apply symdiff_ref.
Qed.

(*Wyckoff*)
Definition Setminus_full {T:Type} (F X Y:Ensemble T) := Intersection X (Setminus F Y).

(*Wyckoff*)
Lemma setminus_full_ref : forall {T:Type} (F X:Ensemble T),
                            Setminus_full F X X = Empty_set _.
intros T F X.
apply Extensionality_Ensembles; split; auto with sets.
red.
intros x h1.
destruct h1 as [x h1a h1b].
destruct h1b; contradiction.
Qed.

(*Wyckoff*)
Lemma setminus_full_empty_inc : 
  forall {T:Type} (F X Y:Ensemble T),
    Setminus_full F X Y = Empty_set _ -> Included (Intersection F X) Y.
intros T F X Y h1.
red.
intros x h2.
apply NNPP. intro h3.
unfold Setminus_full in h1.
destruct h2 as [x h2l h2r].
assert (h4:In (Intersection X (Setminus F Y)) x).
  constructor; auto; constructor; auto.
rewrite h1 in h4.
contradiction.
Qed.

(*Wyckoff*)
Lemma setminus_full_empty_inc' : 
  forall {T:Type} (F X Y:Ensemble T),
    Setminus_full F X Y = Empty_set _ -> Included X F -> 
    Included X Y.
intros T F X Y h1 h2.
pose proof (setminus_full_empty_inc F X Y h1) as h3.
rewrite inclusion_iff_intersection_eq in h2.
rewrite comm_prod_psa in h2.
rewrite h2 in h3.
assumption.
Qed.

(*Wyckoff*)
Definition Symdiff_full {T:Type} (F X Y:Ensemble T) := 
  Union (Setminus_full F X Y) (Setminus_full F Y X).

(*Wyckoff*)
Lemma symdiff_full_ref : forall {T:Type} (F A:Ensemble T),
                        Symdiff_full F A A = Empty_set _.
intros T F A. apply Extensionality_Ensembles; split; auto with sets.
red. intros x h1.
destruct h1 as [x h1 |x h2].
rewrite setminus_full_ref in h1. assumption.
rewrite setminus_full_ref in h2. assumption.
Qed.

(*Wyckoff*)
Lemma symdiff_full_comm : forall {T:Type} (F A B:Ensemble T),
                       Symdiff_full F A B = Symdiff_full F B A.
intros T F A B.
unfold Symdiff_full. auto with sets.
Qed.

(*Wyckoff*)
Lemma symdiff_full_empty_iff_eq : 
  forall {T:Type} (F A B:Ensemble T),
    Included A F -> Included B F ->
    (Symdiff_full F A B = Empty_set _ <-> A = B).
intros T F A B h1 h2. split.
(* -> *)
intro h3.
unfold Symdiff_full in h3.
pose proof (union_eq_empty _ _ h3) as h4.
destruct h4 as [h4l h4r].
pose proof (setminus_full_empty_inc' _ _ _ h4l h1) as h5.
pose proof (setminus_full_empty_inc' _ _ _ h4r h2) as h6.
apply Extensionality_Ensembles; red; split; assumption.
(* <- *)
intro h3.
rewrite h3.
apply symdiff_full_ref.
Qed.


End Symdiff.



Section DeMorgan.
Variable T:Type.

(*Wyckoff*)
Lemma comp_union : forall (M N : Ensemble T), Ensembles.Complement (Union M N) =
  Intersection (Ensembles.Complement M) (Ensembles.Complement N).
intros M N.
apply Extensionality_Ensembles. red. split.
(*left*)
red. intros x h1.
assert (h2: Included M (Union M N)).
  auto with sets.
assert (h3: Included N (Union M N)).
  auto with sets.
split.
pose proof (complement_inclusion _ _ h2) as h4.
auto with sets.
pose proof (complement_inclusion _ _ h3) as h5.
auto with sets.
(*right*)
red. intros x h6.
inversion h6 as [y h7 h8].
unfold Ensembles.Complement in h7. red in h7.
unfold Ensembles.Complement in h8. red in h8.
unfold Ensembles.Complement. red.
intro h9.
inversion h9.
contradiction. contradiction.
Qed.

(*Wyckoff*)
Lemma comp_int : forall (M N : Ensemble T), Ensembles.Complement (Intersection M N) =
  Union (Ensembles.Complement M) (Ensembles.Complement N).
intros M N.
pose proof (Complement_Complement _ M) as h1.
pose proof (Complement_Complement _ N) as h2.
rewrite <- h1 at 1.
rewrite <- h2 at 1.
pose proof (comp_union (Ensembles.Complement M) (Ensembles.Complement N)) as h3.
rewrite <- h3.
apply Complement_Complement.
Qed.
End DeMorgan.

Section Infinite_analogues.
Variable It T:Type.
Variable FI : (IndexedFamily It T).

(*Wyckoff*)
Definition Int_Fam_Ens (P:Ensemble T) (Q:Family T) : Family T := 
  (Im Q (fun (S:Ensemble T) => (Intersection P S))).

(*Wyckoff*)
Definition Union_Fam_Ens (P:Ensemble T) (Q:Family T) : Family T :=
  (Im Q (fun (S:Ensemble T) => (Union P S))).

(*Wyckoff*)
Definition Int_Ind_Ens (P:Ensemble T) (i:It) : (Ensemble T) :=
  Intersection P (FI i).

(*Wyckoff*)
Definition Union_Ind_Ens (P:Ensemble T) (i:It) : (Ensemble T) :=
  Intersection P (FI i).
  
(*Wyckoff*)
Lemma dist_union_infnt_1 : 
  forall (P:Ensemble T) (Q:Family T) , 
    Intersection P (FamilyUnion Q) = FamilyUnion (Int_Fam_Ens P Q).
intros P Q.
apply Extensionality_Ensembles.
unfold Same_set.
split.
(*left*)
unfold Included.
intros x h1.
inversion h1 as [y h2 h3 h4].
inversion h3 as [S z h5 h6 h7].
assert (h8: In (Intersection P S) x).
  auto with sets.
apply family_union_intro with (S := (Intersection P S)).
unfold Int_Fam_Ens.
apply Im_intro with (x := S).
assumption. reflexivity. assumption.
(*right*)
unfold Included.
intros x h8.
inversion h8 as [S y h9 h10 h11].
unfold Int_Fam_Ens in h9.
apply Intersection_intro.
(*inversion h9.*)
inversion h9 as [A h12 B h13 h14].
rewrite h13 in h10.
inversion h10.
assumption.
inversion h9 as [A h12 B h13 h14].
rewrite h13 in h10.
assert (h15: In A x).
  inversion h10. assumption.
apply family_union_intro with (S := A).
assumption. assumption.
Qed.

(*Wyckoff*)
Lemma dist_int_infnt_1 : 
  forall (P:Ensemble T) (Q:Family T), 
    Union P (FamilyIntersection Q) = FamilyIntersection (Union_Fam_Ens P Q).
intros P Q.
apply Extensionality_Ensembles.
unfold Same_set.
split.
(*left*)
unfold Included.
intros x h1.
inversion h1 as [y h2 | y h3].
apply family_intersection_intro.
intros S h4.
unfold Union_Fam_Ens in h4.
inversion h4 as [S' z h5 h6 h7].
assert (h8: In (Union P S') x). auto with sets.
rewrite h6. assumption.
inversion h3 as [z h9 h10].
apply family_intersection_intro.
intros S h11.
unfold Union_Fam_Ens in h11.
inversion h11 as [A h12 B h13 h14].
assert (h15: In A x).
  apply h9.  assumption.
assert (h16: In (Union P A) x).  auto with sets.
rewrite h13. assumption.
(*right*)
unfold Included.
intros x h17.
inversion h17 as [y h18 h19].
unfold Union_Fam_Ens in h18.
assert (h20: forall (S':Ensemble T), (In Q S') -> In (Union P S') x).
  intros S' h50.
  apply h18.
  apply Im_intro with (x := S'). 
  assumption. reflexivity.
assert (h21 : 
    forall S : Ensemble T, In Q S -> In P x \/ In S x).
  intros S h22.
  apply Union_inv.
  apply h20. assumption.
assert (h22: In P x \/ (forall S : Ensemble T, (In Q S -> In S x))).
  apply prop_dis_forall.
  intro S.
  assert (h23: (In Q S -> In P x \/ In S x)).
    apply h21.
  tauto.
case h22.
  auto with sets.
intro h24.
apply Union_intror.
apply family_intersection_intro.
assumption.
Qed.

(*Wyckoff*)
Lemma dist_union_infnt_1_ind : 
  forall (P:Ensemble T), 
    Intersection P (IndexedUnion _ FI) = IndexedUnion _ (Int_Ind_Ens P).
intro P.
rewrite (indexed_to_family_union _ _ FI).
rewrite indexed_to_family_union.
rewrite (dist_union_infnt_1 P (ImageFamily It T FI)).
apply Extensionality_Ensembles. unfold Same_set.
unfold Int_Fam_Ens.
unfold Int_Ind_Ens.
unfold Included.
split.
(*left*)
intros x h1.
inversion h1.
inversion H.
unfold ImageFamily.
apply family_union_intro with (S := S).
unfold ImageFamily in H2.
rewrite H3.
inversion H2.
apply Im_intro with (X := Full_set It) (x := x2).
assumption.
rewrite H6. reflexivity.
assumption.
(*right*)
intros x h20.
unfold ImageFamily in h20.
inversion h20.
apply family_union_intro with (S := S).
inversion H.
rewrite H3.
apply Im_intro with (x := (FI x1)).
unfold ImageFamily.
apply Im_intro with (x := x1).
assumption.
reflexivity.
reflexivity.
assumption.
Qed.
End Infinite_analogues.


(*Section Characteristic. Commented to keep the notation top-level*)

(*Schepler*)
Inductive characteristic_function_abstraction {X:Type} (P:X->Prop) (x:X) : Prop :=
  | intro_characteristic_sat: P x ->
    In (characteristic_function_abstraction P) x.

(*Schepler*)
Definition characteristic_function_to_ensemble {X:Type} (P:X->Prop) : Ensemble X :=
  characteristic_function_abstraction P.

(*Schepler*)
Notation "[ x : X | P ]" :=
  (characteristic_function_to_ensemble (fun x:X => P))
  (x ident).

(*Schepler*)
Lemma characteristic_function_to_ensemble_is_identity:
  forall {X:Type} (P:X->Prop),
    [ x:X | P x ] = P.
Proof.
intros.
apply Extensionality_Ensembles; split; red; intros.
destruct H.
exact H.
constructor.
exact H.
Qed.



Section Finite_Families.
(*Wyckoff*)
Lemma Finite_Finite_Union : forall {T:Type} (F:Family T), 
  (forall (S:(Ensemble T)), In F S -> Finite  S) ->
  Finite F -> Finite (FamilyUnion F).
intros T F h1 h2.
induction h2.
(*Empty*)
pose proof (empty_family_union T) as h3.
rewrite h3.
constructor.
(*Add*)
rename x into S.
rename A into F.
rename H into h3.
rename IHh2 into h4.
pose proof (family_union_add F S) as h5.
rewrite h5.
apply Union_preserves_Finite.
apply h1.
auto with sets.
assert (h6:forall S:Ensemble T, In F S -> Finite S).
  intros S0 h7.
  apply h1.
  auto with sets.
apply (h4 h6).
Qed.

(*Wyckoff*)
(* Singleton Family decomposition of a set *)
Definition SingF {T:Type} (S:Ensemble T) := Im S (@Singleton T).  

(*Wyckoff*)
Lemma union_singF_eq : forall {T:Type} (S:Ensemble T), 
  S = FamilyUnion (SingF S).
intros T S.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros x h1.
assert (h2:In (SingF S) (Singleton x)).
  unfold SingF.
  apply Im_intro with x; trivial.
pose proof (In_singleton _ x) as h3.
apply family_union_intro with (Singleton x); assumption.
(* >= *)
red. intros x h1.

inversion h1 as [S0 ? h2 h3].
unfold SingF in h2.
inversion h2 as [? h4 ? h5].
rewrite h5 in h3.
inversion h3 as [h6].
rewrite <- h6; trivial.
Qed.

(*Wyckoff*)
Lemma fin_fin_singf : forall {T:Type} (S:Ensemble T), Finite S 
  -> Finite (SingF S).
intros.
apply finite_image; assumption.
Qed.

End Finite_Families.

(*End Characteristic.*)
Section CartProd.

(*Wyckoff*)
Definition cart_prod {T U:Type} (A:Ensemble T) (B:Ensemble U) :
  Ensemble (T*U) := [pr:T*U | 
                      In A (fst pr) /\ In B (snd pr)].

(*Wyckoff*)
Lemma in_cart_prod : forall {T U:Type} 
                            (A:Ensemble T) (B:Ensemble U)
                            (x:T) (y:U),
                       In A x -> In B y ->
                       In (cart_prod A B) (x, y).
intros A B x y h1 h2.
constructor.
simpl.
split; assumption.
Qed.

(*Wyckoff*)
Lemma in_cart_prod_comm : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U) (pr:T*U),
    In (cart_prod A B) pr -> In (cart_prod B A) (snd pr, fst pr).
intros T U A B pr h1.
constructor.
simpl.
destruct h1 as [h1]. 
destruct h1; split; assumption.
Qed.
                     
(*Wyckoff*)
Lemma cart_prod_sing : 
  forall {T U:Type} (A:Ensemble T) (y:U), 
    cart_prod A (Singleton y) = 
    Im A (fun x:T => (x, y)).
  intros T U A y.
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red.
  intros pr h1.
  unfold cart_prod in h1.
  destruct h1 as [h1].
  destruct h1 as [h1 h2].
  apply  Im_intro with (fst pr). assumption.
  pose proof (Singleton_inv _ _ _ h2) as h3.
  subst.
  apply surjective_pairing.
  (* >= *)
  red.
  intros pr h1.
  destruct h1 as [x h2 pr].
  subst.
  unfold cart_prod.
  constructor. split. simpl. assumption.
  simpl. constructor.
Qed.

  
(*Wyckoff*)
Lemma cart_prod_sing' : 
  forall {T U:Type} (B:Ensemble U) (x:T), 
    cart_prod (Singleton x) B = 
    Im B (fun y:U => (x, y)).
  intros T U B x.
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red.
  intros pr h1.
  unfold cart_prod in h1.
  destruct h1 as [h1].
  destruct h1 as [h1 h2].
  apply  Im_intro with (snd pr). assumption.
  pose proof (Singleton_inv _ _ _ h1) as h3.
  subst.
  apply surjective_pairing.
  (* >= *)
  red.
  intros pr h1.
  destruct h1 as [z h2 pr].
  subst.
  unfold cart_prod.
  constructor. split. simpl. constructor.
  simpl. assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_sing_rev : 
  forall {T U:Type} (A:Ensemble T) (y:U), 
    A = Im (cart_prod A (Singleton y)) (fun pr:T*U => fst pr).
intros T U A y.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros x h1.
apply Im_intro with (x, y).
unfold cart_prod. constructor. simpl.
split; auto with sets. reflexivity.
(* >= *)
red. 
intros x h1.
destruct h1 as [pr h1]; subst.
destruct h1 as [h1].
destruct h1; assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_sing_rev' : 
  forall {T U:Type} (B:Ensemble U) (x:T), 
    B = Im (cart_prod (Singleton x) B) (fun pr:T*U => snd pr).
intros T U B x.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros y h1.
apply Im_intro with (x, y).
unfold cart_prod. constructor. simpl.
split; auto with sets. reflexivity.
(* >= *)
red. 
intros y h1.
destruct h1 as [pr h1]; subst.
destruct h1 as [h1].
destruct h1; assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_sing_fin : forall {T U:Type} (A:Ensemble T) (y:U),
                             Finite A ->
                             Finite (cart_prod A (Singleton y)).
intros T U A y h1.
pose proof (cart_prod_sing A y) as h2.
rewrite h2.
apply finite_image.
assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_sing_fin' : forall {T U:Type} (B:Ensemble U) (x:T),
                             Finite B ->
                             Finite (cart_prod (Singleton x) B).
intros T U B x h1.
pose proof (cart_prod_sing' B x) as h2.
rewrite h2.
apply finite_image.
assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_sing_fin_rev : forall {T U:Type} (A:Ensemble T) (y:U),
                                 Finite (cart_prod A (Singleton y)) ->                            
                                 Finite A.
                            
intros T U A y h1.
pose proof (cart_prod_sing_rev A y) as h2.
rewrite h2.
apply finite_image.
assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_sing_fin_rev' : forall {T U:Type} (B:Ensemble U) (x:T),
                                 Finite (cart_prod (Singleton x) B) ->                            
                                 Finite B.
                            
intros T U B x h1.
pose proof (cart_prod_sing_rev' B x) as h2.
rewrite h2.
apply finite_image.
assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_inc_add1 : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U) (x:T),
    Included (cart_prod A B) (cart_prod (Add A x) B).
intros T U A B x.
red. intros pr h1.
destruct h1 as [h1].
destruct h1 as [h1l h1r].
constructor.
split; try assumption.
left. assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_inc_add2 : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U) (y:U),
    Included (cart_prod A B) (cart_prod A (Add B y)).
intros T U A B y.
red. intros pr h1.
destruct h1 as [h1].
destruct h1 as [h1l h1r].
constructor.
split; try assumption.
left. assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_eq : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U), 
    cart_prod A B =
    FamilyUnion 
      [S:Ensemble (T*U) | 
       exists y:U,
                In B y /\ S = cart_prod A (Singleton y)].
intros T U A B.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros pr h1.
destruct h1 as [h1l h1r].
destruct h1l as [h1ll h1lr].
assert (h2:In (cart_prod A (Singleton (snd pr))) pr).
  rewrite cart_prod_sing.
  apply Im_intro with (fst pr). assumption.
  apply surjective_pairing.
  apply family_union_intro with 
  (cart_prod A (Singleton (snd pr))).
  constructor.
  exists (snd pr). 
  split; [assumption | reflexivity]. assumption.
(* >= *)
red. intros pr h1.
destruct h1 as [S pr h1 h2].
destruct h1 as [h1].
destruct h1 as [y h1].
destruct h1 as [h1l h1r].
subst.
rewrite cart_prod_sing in h2.
destruct h2 as [x].
subst.
constructor; split; assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_eq' : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U), 
    cart_prod A B =
    FamilyUnion 
      [S:Ensemble (T*U) | 
       exists x:T,
                In A x /\ S = cart_prod (Singleton x) B].

intros T U A B.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros pr h1.
destruct h1 as [h1l h1r].
destruct h1l as [h1ll h1lr].
assert (h2:In (cart_prod (Singleton (fst pr)) B) pr).
  rewrite cart_prod_sing'.
  apply Im_intro with (snd pr). assumption.
  apply surjective_pairing.
  apply family_union_intro with 
  (cart_prod (Singleton (fst pr)) B).
  constructor.
  exists (fst pr). 
  split; [assumption | reflexivity]. assumption.
(* >= *)
red. intros pr h1.
destruct h1 as [S pr h1 h2].
destruct h1 as [h1].
destruct h1 as [y h1].
destruct h1 as [h1l h1r].
subst.
rewrite cart_prod_sing' in h2.
destruct h2 as [x].
subst.
constructor; split; assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_empty : forall {T U:Type} (B:Ensemble U),
                          cart_prod (Empty_set T) B =
                          Empty_set _.
intros T U B.
apply Extensionality_Ensembles.
red. split.
red.
intros pr h1.
unfold cart_prod in h1.
destruct h1 as [h1].
destruct h1; contradiction.
auto with sets.
Qed.

(*Wyckoff*)
Lemma cart_prod_empty_empty : forall {T U:Type} (B:Ensemble U),
                          cart_prod (Empty_set T) B =
                          cart_prod (Empty_set _) (Empty_set _).
intros T U B.
rewrite cart_prod_empty.
rewrite <- (cart_prod_empty (Empty_set _)).
reflexivity.
Qed.

(*Wyckoff*)
Lemma cart_prod_empty' : forall {T U:Type} (A:Ensemble T),
                          cart_prod A (Empty_set U)=
                          Empty_set _.
intros T U A.
apply Extensionality_Ensembles.
red. split.
red.
intros pr h1.
unfold cart_prod in h1.
destruct h1 as [h1].
destruct h1; contradiction.
auto with sets.
Qed.

(*Wyckoff*)
Lemma cart_prod_empty_rev : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U),
    cart_prod A B = Empty_set _ -> 
    (A = Empty_set _ \/ B = Empty_set _).
intros T U A B h1.
apply NNPP.
intro h2.
pose proof (not_or_and _ _ h2) as h3.
destruct h3 as [h3 h4].
pose proof (not_empty_Inhabited _ _ h3) as h5.
pose proof (not_empty_Inhabited _ _ h4) as h6.
destruct h5 as [a h5].
destruct h6 as [b h6].
assert (h7:In (cart_prod A B) (a, b)).
  constructor. simpl. split; assumption.
rewrite h1 in h7.
contradiction.
Qed.

(*Wyckoff*)
Lemma cart_prod_empty_empty' : forall {T U:Type} (A:Ensemble U),
                          cart_prod A (Empty_set U) =
                          cart_prod (Empty_set _) (Empty_set _).
intros T U B.
rewrite cart_prod_empty'.
rewrite <- (cart_prod_empty' (Empty_set _)).
reflexivity.
Qed.

(*Wyckoff*)
Lemma add_ex_family : 
  forall {T U:Type} (A:Ensemble U) (a:U)
         (f:U -> Ensemble T),
    [S:Ensemble T | exists u:U, Ensembles.In (Add A a) u /\ S = f u] =
    Add [S:Ensemble T | exists u:U, Ensembles.In A u /\ S = f u]
        (f a).
  intros T U A a f.
    apply Extensionality_Ensembles.
    red. split.
    (* <= *)
    red.
    intros S h6.
    destruct h6 as [h6].
    destruct h6 as [y h6].
    destruct h6 as [h6l h6r].
    destruct h6l as [y h6la | y h6lb].
    left.
    constructor. exists y. split; assumption.
    destruct h6lb.
    right. rewrite h6r.
    constructor.
    (* >= *)
    red.
    intros S h6.
    destruct h6 as [S h6 | S h7'].
    destruct h6 as [h6].
    destruct h6 as [y h6].
    destruct h6 as [h6a h6b].
    constructor.
    exists y.
    split. left. assumption. assumption.
    destruct h7'.
    constructor.
    exists a. split.
    right. constructor. reflexivity.
Qed.

(*Wyckoff*)
Lemma cart_prod_fin : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U), 
    Finite A -> Finite B -> Finite (cart_prod A B).
  intros T U A B h1.
  induction h1.
    intros h2. rewrite cart_prod_empty.
    constructor.
  intro h2.
  rewrite cart_prod_eq.
  apply Finite_Finite_Union.
  intros S h3.
  destruct h3 as [h3].
  destruct h3 as [y h3].
  destruct h3 as [h3l h3r].
  rewrite h3r.
  apply cart_prod_sing_fin.
  apply Add_preserves_Finite. assumption.
  induction h2 as [|B h3 h4 z h5].
  assert (h3: [S : Ensemble (T * U)
     | exists y : U, In (Empty_set U) y /\ S = cart_prod (Add A x) (Singleton y)] = Empty_set _ ).
  apply Extensionality_Ensembles.
  red. split.
  red.
  intros S h2. destruct h2 as [h2].
  destruct h2 as [y h2].
  destruct h2; contradiction.
  auto with sets.
  rewrite h3.
  constructor.
  pose proof (Add_preserves_Finite _ B z h3) as h0.
  specialize (IHh1 h0).
  rewrite cart_prod_eq in IHh1.
  assert (hs : Included (FamilyUnion [S:Ensemble (T*U) |
                            exists y:U, In B y /\ 
                                        S = cart_prod A (Singleton y)]) (FamilyUnion
              [S : Ensemble (T * U)
              | exists y : U, In (Add B z) y /\ S = cart_prod A (Singleton y)])).
    red. intros pr h6.
    destruct h6 as [S pr h7].
    apply family_union_intro with S.
    destruct h7 as [h7].
    destruct h7 as [y h7].
    constructor. exists y.
    split.
    left. tauto. tauto. assumption.
  pose proof (Finite_downward_closed _ _ IHh1 _ hs) as h7.
  rewrite <- cart_prod_eq in h7.
  assert (h8:Finite B -> Finite (cart_prod A B)).
    tauto.
  specialize (h4 h8).
  pose proof (@add_ex_family).
  pose proof (add_ex_family B z (fun u:U => cart_prod (Add A x)
                                                      (Singleton u))) as h6.
  simpl in h6.
  rewrite h6.
  constructor. assumption.
  intro h9.
  destruct h9 as [h9].
  destruct h9 as [a h9].
  destruct h9 as [h9a h9b].
  assert (h10:In (cart_prod (Add A x) (Singleton z)) (x, z)).
    constructor.
    simpl. split.
    right. constructor.
    constructor.
  rewrite h9b in h10.
  destruct h10 as [h10].
  destruct h10 as [h10a h10b].
  simpl in h10b.
  destruct h10b.
  contradiction.
Qed.

(*Wyckoff*)
Lemma cart_prod_fin_rev1 : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U), 
    Finite (cart_prod A B) -> Inhabited B ->
    Finite A.
intros T U A B h1 h2.

dependent induction h1.
destruct h2 as [y h2].

assert (h3:A = Empty_set _).
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red.
  intros z h1.
  assert (h3:In (cart_prod A B) (z, y)). 
    constructor. simpl. split; assumption.
  rewrite <- x in h3.
  contradiction.
  (* >= *)
  auto with sets.
rewrite h3. constructor.
pose proof (Add_preserves_Finite _ _ x0 h1) as h3.
rewrite x in h3.
rewrite cart_prod_eq in h3.
assert (h4:Included (cart_prod A (Singleton (snd x0)))
(FamilyUnion
            [S : Ensemble (T * U)
            | exists y : U, In B y /\ S = cart_prod A (Singleton y)])).
  red.
  intros pr h4.
  apply family_union_intro with (cart_prod A (Singleton (snd x0))).
  constructor.
  exists (snd x0).
  split.
  pose proof (Add_intro2 _ A0 x0) as h5.
  rewrite x in h5.
  destruct h5 as [h5]. destruct h5; assumption.
  reflexivity.
  assumption.
pose proof (Finite_downward_closed _ _ h3 _ h4) as h5.
apply cart_prod_sing_fin_rev with (snd x0).
assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_fin_rev2 : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U), 
    Finite (cart_prod A B) -> Inhabited A ->
    Finite B.
intros T U A B h1 h2.

dependent induction h1.
destruct h2 as [a h2].

assert (h3:B = Empty_set _).
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red.
  intros b h1.
  assert (h3:In (cart_prod A B) (a, b)). 
    constructor. simpl. split; assumption.
  rewrite <- x in h3.
  contradiction.
  (* >= *)
  auto with sets.
rewrite h3. constructor.
pose proof (Add_preserves_Finite _ _ x0 h1) as h3.
rewrite x in h3.
rewrite cart_prod_eq' in h3.
assert (h4:Included (cart_prod (Singleton (fst x0)) B)
(FamilyUnion
            [S : Ensemble (T * U)
            | exists x1 : T, In A x1 /\ S = cart_prod (Singleton x1) B])).
  red.
  intros pr h4.
  apply family_union_intro with (cart_prod (Singleton (fst x0)) B).
  constructor.
  exists (fst x0).
  split.
  pose proof (Add_intro2 _ A0 x0) as h5.
  rewrite x in h5.
  destruct h5 as [h5]. destruct h5; assumption.
  reflexivity.
  assumption.
pose proof (Finite_downward_closed _ _ h3 _ h4) as h5.
apply cart_prod_sing_fin_rev' with (fst x0).
assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_fin_rev : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U), 
    Finite (cart_prod A B) -> Inhabited (cart_prod A B) ->
    Finite A /\ Finite B.
intros T U A B h1 h2.
destruct h2 as [pr h2].
destruct h2 as [h2].
destruct h2 as [h2l h2r].
pose proof (Inhabited_intro _ _ _ h2l) as h3.
pose proof (Inhabited_intro _ _ _ h2r) as h4.
split.
apply cart_prod_fin_rev1 with B; assumption.
apply cart_prod_fin_rev2 with A; assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_fin_rev_or : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U),
    Finite (cart_prod A B) ->
    Finite A \/ Finite B.
intros T U A B h1.
destruct (classic (Inhabited A)) as [h2 | h3].
right.
apply cart_prod_fin_rev2 with A; assumption.
left.
assert (h4:A = Empty_set _).
  apply Extensionality_Ensembles;split. 
  red. intros x h4. 
  pose proof (Inhabited_intro _ A _ h4).
  contradiction. 
  auto with sets. 
rewrite h4.
constructor.
Qed.

(* See TypeUtilties for 
Lemma cart_prod_fin_comm : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U),
          Finite (cart_prod A B) ->
          Finite (cart_prod B A).*)

(*Wyckoff*) 
Lemma cart_prod_proj1_surj : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U),
    let f := (fun pr:T*U => fst pr) in
    Inhabited B -> Im (cart_prod A B) f = A.
intros T U A B f h1.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros y h2.
destruct h2 as [x h2].
subst.
destruct h2 as [h2].
destruct h2 as [h2l h2r].
unfold f. simpl.
assumption.
(* >= *)
red.
intros x h2.
unfold f.
destruct h1 as [y h1].
apply Im_intro with (x, y).
split; simpl; split; assumption.
simpl. reflexivity.
Qed.

(*Wyckoff*)
Lemma cart_prod_proj2_surj : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U),
    let f := (fun pr:T*U => snd pr) in
    Inhabited A -> Im (cart_prod A B) f = B.
intros T U A B f h1.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros y h2.
destruct h2 as [x h2].
subst.
destruct h2 as [h2].
destruct h2 as [h2l h2r].
unfold f. simpl.
assumption.
(* >= *)
red.
intros y h2.
unfold f.
destruct h1 as [x h1].
apply Im_intro with (x, y).
split; simpl; split; assumption.
simpl. reflexivity.
Qed.


(*Wyckoff*)
Lemma cart_prod_im : 
  forall {T U V W:Type} 
         (A:Ensemble T) (B:Ensemble U) (f:T->V) (g:U->W),
         cart_prod (Im A f) (Im B g) = 
         Im (cart_prod A B) (fun pr => (f (fst pr), g (snd pr))).
intros T U V W A B f g.
apply Extensionality_Ensembles.

red. split.
(* <= *)
red.
intros pr h1.
inversion h1 as [h2]. clear h1.
destruct h2 as [h2l h2r].
inversion h2l as [x h3 v h4 h5]. clear h2l.
inversion h2r as [y h6 w h7 h8]. clear h2r.
rewrite (surjective_pairing pr).
subst. rewrite h4. rewrite h7.
apply Im_intro with (x, y). constructor; auto.
simpl. reflexivity.
(* >= *)
red. intros pr h1.
inversion h1 as [prx  h3 pry  h4 h5]. clear h1. 
subst.
rewrite (surjective_pairing prx) in h3.
destruct h3 as [h3].
simpl in h3.
destruct h3 as [h3l h3r].
constructor. simpl. split.
apply Im_intro with (fst prx).  assumption. reflexivity.
apply Im_intro with (snd prx).  assumption. reflexivity.
Qed.




End CartProd.



Section SurjectiveEq.

(*Wyckoff*)
Lemma surj_eq : forall {T U:Type} (f:T->U), surjective f <-> Im (Full_set T) f = 
  Full_set U.
intros T U f. 
split.
(* -> *)
intro h1.
apply Extensionality_Ensembles.
red. split.
  (* <= *)
  red.
  intros x ?. constructor.
  (* >= *)
  red.
  intros u ?.
  red in h1.
  pose proof (h1 u) as h2.
  destruct h2 as [t].
  apply Im_intro with t.
  constructor.
  symmetry; assumption.
(* <- *)
intro h1.
red.
intro u.
pose proof (Full_intro _ u) as h2.
rewrite <- h1 in h2.
inversion h2 as [t].
exists t. symmetry. assumption.
Qed.

End SurjectiveEq.

Section RandomFacts.

(*Wyckoff*)
Lemma not_inhabited_empty : forall {T:Type} (S:Ensemble T),
                              ~ Inhabited S -> S = Empty_set _.
intros T S h1. 
apply Extensionality_Ensembles.
red. split.
red.
intros x h2.
pose proof (Inhabited_intro _ _ _ h2). contradiction.
auto with sets.
Qed.

(*Wyckoff*)
Lemma full_sig : forall {T:Type} (S:Ensemble T), 
  S = Im (Full_set {x:T | In S x}) (@proj1_sig _ _).
intros T S.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros x h1.
apply Im_intro with (exist _ x h1).
constructor. simpl. reflexivity.
(* >= *)
red. intros x h1.
inversion h1 as [x' h2 ? h3 h4].
rewrite h3.
apply proj2_sig.
Qed.

(*Wyckoff*)
Lemma feq_im_aux : forall {T U:Type} (A B:Ensemble T) (f:T->U),
                 A = B -> Included (Im A f) (Im B f).
intros T U A B f h1.
red.
intros y h2.
destruct h2; subst.
apply Im_intro with x; auto.
Qed.

(*Wyckoff*)
Lemma feq_im : forall {T U:Type} (A B:Ensemble T) (f:T->U),
                 A = B -> Im A f = Im B f.
intros T U A B f h1.
apply Extensionality_Ensembles; red; split; apply feq_im_aux; auto.
Qed.

(*Wyckoff*)
Lemma image_union : forall {T U:Type} (A B:Ensemble T) (f:T->U),
                      Im (Union A B) f = Union (Im A f) (Im B f).
intros T U A B f.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros x h1.
destruct h1 as [x h1 y h3].
subst.
destruct h1 as [x h1a | x h1b].
left. apply Im_intro with x; auto.

right. apply Im_intro with x; auto.
(* >= *)
red. 
intros x h1. 
destruct h1 as [x h1a | x h1b].
destruct h1a as [x h1a y h2]. subst.
apply Im_intro with x. left; auto. reflexivity.
destruct h1b as [x h1b y h2]. subst.
apply Im_intro with x; try auto; right; auto.
Qed.                       

(*Wyckoff*)
Lemma excl_middle_sat : forall {T:Type} (P:T->Prop),
                          Full_set T = Union [x:T | P x] [x:T | ~ (P x)].
intros T P. apply Extensionality_Ensembles.
red. split.
red.
intros x h1.
destruct (classic (P x)) as [h2 | h3].
left. constructor. assumption.
right. constructor. assumption.
red. intros; constructor.
Qed.

(*Wyckoff*)
Lemma excl_middle_sat' : forall {T:Type} (S:Ensemble T)
                         (P:T->Prop),
                           S = Union [x:T | In S x /\ P x]
                                     [x:T | In S x /\ ~ (P x)].
intros T S P.
apply Extensionality_Ensembles.
red. split.
red.
intros x h1.
destruct (classic (P x)) as [h2 | h3].
left. constructor. split; auto.
right. constructor. split; auto.
red. intros x h1.
destruct h1 as [x h1l | x h1r].
destruct h1l as [h1l]. destruct h1l; assumption.
destruct h1r as [h1r]. destruct h1r; assumption.
Qed.

(*Wyckoff*)
Lemma im_im : 
  forall {T U V:Type} (A:Ensemble T) (f:T->U) (g:U->V),
    Im (Im A f) g = Im A (fun x => g (f x)).
intros T U V A f g.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros v h1.
inversion h1 as [x h2 v' h3 h4].  subst.
destruct h2 as [x h2 u h3]. subst.
apply Im_intro with x. assumption. reflexivity.
(* >= *)
red. intros v h1.
destruct h1 as [x h2 v' h3]. subst.
apply Im_intro with (f x). apply Im_intro with x; auto.
reflexivity.
Qed.

(*Wyckoff*)
Lemma union_setminus : forall {T:Type} (A X:Ensemble T),
                         Union A (Setminus X A) = Union A X.
intros T A X.
rewrite setminus_int_complement.
rewrite dist_prod_psa.
rewrite excl_middle_full.
rewrite intersection_full_set.
reflexivity.
Qed.

(*Wyckoff*)
Lemma union_setminus_inc : forall {T:Type} (A X:Ensemble T),
                          Included A X ->
                          Union A (Setminus X A) = X.
intros T A X h1.
rewrite union_setminus.
rewrite <- inclusion_iff_union.
assumption.
Qed.

(*Wyckoff*)
Lemma subtract_nin : forall {T:Type} (A:Ensemble T) (x:T),
                       ~In A x -> Subtract A x = A.
intros T A x h1.
apply Extensionality_Ensembles.
red. split. red. intros y h2. destruct h2; auto.
red. intros y h2. constructor; auto.
intro h3.
destruct h3; subst.
contradiction.
Qed.

(*Wyckoff*)
Lemma incl_subtract : 
  forall {T:Type} (A B:Ensemble T) (x:T), 
    Included A B -> Included (Subtract A x) (Subtract B x).
intros T A B x h1.
red. intros y h2.
destruct h2 as [h2 h3].
constructor; auto with sets.
Qed.

(*Wyckoff*)
Lemma atom_sing : 
  forall {T:Type} (E A:Ensemble T), 
    Finite E -> Included A E -> A <> Empty_set _ ->
    ((forall F:Ensemble T, Included F E ->
                          Intersection F A = A \/
                          Intersection F A = Empty_set _) <->
    (exists (a:T), In E a /\ A = Singleton a)).
intros T E A h1. revert A.
induction h1 as [|E h2 h3 x h4].
intros A h1 h2.
assert (h3:A = Empty_set _). auto with sets.
contradiction.
intros A h1 h5. 
split. 
(* -> *)
intro h6.
pose proof (incl_subtract A (Add E x) x h1) as h7.
rewrite sub_add_compat_nin in h7; auto.
destruct (classic (In A x)) as [h8 | h9].
exists x. split. right. constructor. 
pose proof (Inhabited_intro _ _ _ h8) as h9.

assert (h10:Included (Singleton x) (Add E x)).
  red. intros x' h11. destruct h11. right. constructor.
specialize (h6 (Singleton x) h10).
destruct h6 as [h6l | h6r].
assert (h11:Intersection (Singleton x) A = Singleton x).
  apply Extensionality_Ensembles.
  red. split. auto with sets.
  red. intros x' h11. destruct h11. constructor; auto.
  constructor.
rewrite h11 in  h6l.
rewrite h6l. reflexivity.
assert (h11:In (Intersection (Singleton x) A) x).
  constructor; auto. constructor.
rewrite h6r in h11. contradiction.
pose proof (incl_subtract A (Add E x) x h1) as h10.
rewrite subtract_nin in h10; auto.
rewrite sub_add_compat_nin in h10; auto.
specialize (h3 _ h10 h5).
assert (h11:(forall F : Ensemble T,
        Included F E ->
        Intersection F A = A \/ Intersection F A = Empty_set T)).
  intros F h12. 
  assert (h13:Included F (Add E x)). red. intros; auto with sets.
  apply h6; auto.
rewrite h3 in h11.
destruct h11 as [a h11].
destruct h11 as [h11l h11r].
exists a. split; auto. left. assumption.
intros h6.
destruct h6 as [a h6].
destruct h6 as [h6l h6r].
intros F h7.
subst.
destruct (classic (In F a)) as [h8 | h9].
left. apply Extensionality_Ensembles. red. split.  auto with sets.
red. intros a' h9. destruct h9; subst. constructor; auto.
constructor.
right.
apply Extensionality_Ensembles.
red. split; auto with sets.
red. intros a' h10. destruct h10 as [a' h10l h10r].
destruct h10r; subst.
contradiction.
Qed.

(*Wyckoff*)
Inductive inc_sig {T:Type} (A B:Ensemble T)
          (pfi:Included A B) : Ensemble {x:T|In B x} :=
  inc_sig_intro : forall (x:T) (pfa:In A x),
                    In (inc_sig A B pfi) 
                       (exist _ _ (pfi _ pfa)).

(*Wyckoff*)
Lemma inc_sig_iff :   
  forall {T:Type} (A B:Ensemble T) (pf:Included A B)
         x, In (inc_sig A B pf) x <->
            In A (proj1_sig x).
intros T A B h0 x.
split. intros h1.
destruct h1. simpl. assumption.
intro h1. 
pose proof (inc_sig_intro _ _ h0 _ h1) as h2.
assert (h3:x = (exist (In B) (proj1_sig x) (h0 (proj1_sig x) h1))). 
  rewrite unfold_sig at 1. apply proj1_sig_injective. simpl.
  reflexivity.
rewrite <- h3 in h2.
assumption.
Qed.

(*Wyckoff*)
Definition subset_sig {T:Type} (A B:Ensemble T)
  (pf:Included A B) : Ensemble {x:T | In B x}.
red in pf.
refine (Im (Full_set {x:T | In A x}) (fun x => exist _ _ (pf (proj1_sig x) (proj2_sig x)))). 
Defined.


(*Wyckoff*)
Lemma subset_sig_compat : 
  forall {T:Type} (A B:Ensemble T) (pf:Included A B)
         x, In (subset_sig A B pf) x <->
                In A (proj1_sig x).
intros T A B h0 x.
split.
intro h1.
destruct h1 as [x h2 y h4]. subst. simpl.
inversion h2. subst. apply proj2_sig.
intro h1.
unfold subset_sig.
apply Im_intro with  (exist _ _ h1).
constructor. simpl.
apply proj1_sig_injective. simpl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma inc_sig_eq_subset_sig : forall {T:Type} (A B:Ensemble T)
  (pf:Included A B), inc_sig A B pf = subset_sig A B pf.
intros T A B pf.
apply Extensionality_Ensembles.
red. split.
red.
intros x h1.
destruct h1 as [x h1].
unfold subset_sig.
apply Im_intro with (exist _ _ h1).
constructor. apply proj1_sig_injective. simpl. reflexivity.
red. 
intros x h1.
destruct h1 as [x h1]. subst.
constructor.
Qed.


(*Wyckoff*)
Lemma full_empty_sig_empty : forall (T:Type),
                                     Full_set {x:T | In (Empty_set T) x} = Empty_set _.
intros T.
apply Extensionality_Ensembles.
red. split.
red.
intros x h1.
pose proof (proj2_sig x) as h2. simpl in h2.
contradiction.
auto with sets.
Qed.

(*Wyckoff*)
Lemma finite_image_rev_inj : 
  forall {T U:Type} (A:Ensemble T) (f:T->U), 
    injective f -> Finite (Im A f) ->
    Finite A.
intros T U A f h1 h2. 
apply NNPP.
intro h3. 
contradict h2.
intro h4.
pose proof (Pigeonhole_bis _ _ _ _ h3 h4).
contradiction.
Qed.

(*Wyckoff*)
Definition inv_im {T U:Type} (B:Ensemble U) (f:T->U) :=
  [x : T | In B (f x)].


(*Wyckoff*)
Lemma im_inv_im : 
  forall {T U:Type} (f:T->U) (B:Ensemble U),
    Im (inv_im B f) f = Intersection B (Im (Full_set T) f).
intros T U  f B.
apply Extensionality_Ensembles.
unfold inv_im.
red. split.
(* <= *)
red.
intros b h1. 
inversion h1 as [a h2 b' h3]. subst.
constructor. destruct h2. assumption.
apply Im_intro with a. constructor. reflexivity.
red.
intros u h1.
inversion h1 as [u' h2 h3]. subst.
inversion h3 as [x h4 y h5 h6]. subst.
apply Im_intro with x.
constructor. assumption.
reflexivity.
Qed.



End RandomFacts.

Arguments FamilyUnion [T] F _.
Arguments FamilyIntersection [T] F _.
Arguments IndexedUnion [It] [T] FI _.
Arguments IndexedIntersection [It] [T] FI _.
Arguments assoc_sum_psa [Xt] N M P.
Arguments assoc_prod_psa [Xt] N M P.
Arguments comm_sum_psa [Xt] N M. 
Arguments comm_prod_psa [Xt] N M.
Arguments abs_sum_psa [Xt] N M. 
Arguments abs_prod_psa [Xt] N M.
Arguments dist_sum_psa [Xt] N M P.
Arguments dist_prod_psa [Xt] N M P.
Arguments comp_sum_psa [Xt] N. 
Arguments comp_prod_psa [Xt] N.

